#![cfg_attr(not(debug_assertions), windows_subsystem = "windows")]

pub mod components;
pub mod constants;
mod map;
pub mod resources;
mod sidebar;
mod sound;
mod sound_effects;
pub mod systems;
pub mod vga_color;

use bracket_lib::prelude::*;
use components::{
    monster::{color_for_kind, damage_for_kind, glyphs_for_kind, ticks_for_kind, MonsterKind},
    Monster, Player, Position, Renderable,
};
use constants::*;
use map::{Map, TileType};
use resources::{Clock, LevelNumber, ShowDebugInfo, Stats};
use sound::SoundSystem;
use sound_effects::SoundEffects;
use specs::prelude::*;
use std::time::{Duration, Instant};
use systems::MonsterMotion;
use vga_color as vga;

struct State {
    ecs: World,
    sound_system: SoundSystem,
}

fn try_move_player(delta_x: i32, delta_y: i32, ecs: &mut World, sound_system: &mut SoundSystem) {
    let entities = ecs.entities();
    let mut positions = ecs.write_storage::<Position>();
    let mut players = ecs.write_storage::<Player>();
    let mut map = ecs.write_resource::<Map>();
    let mut stats = ecs.write_resource::<Stats>();

    for (player, pos) in (&mut players, &mut positions).join() {
        let now = Instant::now();
        if now - player.last_moved > Duration::from_secs_f32(PLAYER_STEP_PERIOD) {
            let destination = Point {
                x: pos.x + delta_x,
                y: pos.y + delta_y,
            };

            let mut sound_effects = ecs.fetch_mut::<SoundEffects>();

            if map.in_bounds(destination) {
                if map.is_solid(destination) {
                    sound_system.play_sound(sound_effects.blocked.clone());
                } else {
                    if let TileType::Monster(monster) = map.get_tile_at(destination) {
                        if let Some(monster_component) =
                            ecs.read_component::<Monster>().get(monster)
                        {
                            if stats.gems > 0 {
                                stats.gems -= damage_for_kind(monster_component.kind);
                            }
                            let _ = entities.delete(monster);
                        }
                    }
                    map.set_tile_at(Point { x: pos.x, y: pos.y }, TileType::Floor);
                    map.set_tile_at(destination, TileType::Player);

                    pos.x = destination.x;
                    pos.y = destination.y;

                    let mut player_pos = ecs.write_resource::<Point>();
                    player_pos.x = pos.x;
                    player_pos.y = pos.y;

                    ecs.write_resource::<Clock>().force_tick();

                    sound_system.play_sound(sound_effects.step.clone());
                }
            } else {
                let static_sound = sound_effects.get_new_static_effect(sound_system);
                sound_system.play_sound(static_sound);
            }

            player.last_moved = now;
        }
    }
}

fn player_input(gs: &mut State, ctx: &mut BTerm) {
    // Player movement
    match ctx.key {
        None => {} // Nothing happened
        Some(key) => match key {
            VirtualKeyCode::Left | VirtualKeyCode::J => {
                try_move_player(-1, 0, &mut gs.ecs, &mut gs.sound_system);
            }
            VirtualKeyCode::U | VirtualKeyCode::Home => {
                try_move_player(-1, -1, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::Up | VirtualKeyCode::I => {
                try_move_player(0, -1, &mut gs.ecs, &mut gs.sound_system);
            }
            VirtualKeyCode::O | VirtualKeyCode::PageUp => {
                try_move_player(1, -1, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::Right | VirtualKeyCode::K => {
                try_move_player(1, 0, &mut gs.ecs, &mut gs.sound_system);
            }
            VirtualKeyCode::Comma | VirtualKeyCode::PageDown => {
                try_move_player(1, 1, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::Down | VirtualKeyCode::M => {
                try_move_player(0, 1, &mut gs.ecs, &mut gs.sound_system);
            }
            VirtualKeyCode::N | VirtualKeyCode::End => {
                try_move_player(-1, 1, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::S => {
                let sound_effects = gs.ecs.fetch::<SoundEffects>();
                gs.sound_system.play_sound(sound_effects.pickup.clone());
            }
            VirtualKeyCode::D => {
                let mut show_debug_info = gs.ecs.write_resource::<ShowDebugInfo>();
                show_debug_info.0 = !show_debug_info.0;
            }
            VirtualKeyCode::Escape => {
                ctx.quit();
            }
            _ => {
                let sound_effects = gs.ecs.fetch::<SoundEffects>();
                gs.sound_system.play_sound(sound_effects.bad_key.clone());
            }
        },
    }
}

impl GameState for State {
    fn tick(&mut self, ctx: &mut BTerm) {
        ctx.cls();

        player_input(self, ctx);
        self.run_systems();

        let map = self.ecs.fetch::<Map>();
        map.draw(ctx);

        let positions = self.ecs.read_storage::<Position>();
        let renderables = self.ecs.read_storage::<Renderable>();

        for (pos, render) in (&positions, &renderables).join() {
            ctx.set(
                pos.x + MAP_X as i32,
                pos.y + MAP_Y as i32,
                render.fg,
                render.bg,
                render.glyph,
            );
        }

        sidebar::draw(&self.ecs, ctx);
    }
}

impl State {
    fn run_systems(&mut self) {
        let mut mm = MonsterMotion {};
        mm.run_now(&self.ecs);
        self.ecs.maintain();
        self.ecs.write_resource::<Clock>().try_tick();
    }
}

fn main() -> BError {
    let context = BTermBuilder::simple(80, 25)?
        .with_fps_cap(60.0)
        .with_title("Kroz")
        .with_tile_dimensions(8, 16)
        .build()?;

    let mut ss = SoundSystem::new();
    let sound_effects = SoundEffects::new(&ss);
    ss.play_sound(sound_effects.startup.clone());

    let mut gs = State {
        ecs: World::new(),
        sound_system: ss,
    };

    gs.ecs.insert(LevelNumber(0));
    gs.ecs.insert(ShowDebugInfo(false));
    gs.ecs.insert(Clock {
        last_ticked: Instant::now(),
        has_ticked: false,
        ticks: 0,
    });
    gs.ecs.insert(sound_effects);
    gs.ecs.insert(Stats {
        score: 1290,
        gems: 14,
        whips: 7,
        teleports: 0,
        keys: 0,
    });

    gs.ecs.register::<Position>();
    gs.ecs.register::<Renderable>();
    gs.ecs.register::<Monster>();
    gs.ecs.register::<Player>();

    let mut map = Map::new();
    let mut rng = RandomNumberGenerator::new();
    for (i, tile) in &mut map.get_tiles().iter().enumerate() {
        if rng.roll_dice(1, 16) < 2 && *tile == TileType::Floor {
            let position = map.index_to_point2d(i);
            let kind = MonsterKind::Slow;
            gs.ecs
                .create_entity()
                .with(Position {
                    x: position.x,
                    y: position.y,
                })
                .with(Renderable {
                    glyph: *rng.random_slice_entry(&glyphs_for_kind(kind)).unwrap(),
                    fg: color_for_kind(kind),
                    bg: RGB::named(vga::BLACK),
                })
                .with(Monster {
                    kind,
                    ticks_until_move: ticks_for_kind(kind),
                })
                .build();
        }
    }

    {
        let entities = gs.ecs.entities();
        let positions = gs.ecs.read_storage::<Position>();
        let monsters = gs.ecs.read_storage::<Monster>();

        for (entity, _monster, pos) in (&entities, &monsters, &positions).join() {
            map.set_tile_at(Point { x: pos.x, y: pos.y }, TileType::Monster(entity));
        }
    }

    let player_start_pos = Point { x: 40, y: 22 };

    gs.ecs
        .create_entity()
        .with(Position {
            x: player_start_pos.x,
            y: player_start_pos.y,
        })
        .with(Renderable {
            glyph: to_cp437('☻'),
            fg: RGB::named(vga::YELLOW_BRIGHT),
            bg: RGB::named(vga::BLACK),
        })
        .with(Player {
            last_moved: Instant::now(),
        })
        .build();

    map.set_tile_at(player_start_pos, TileType::Player);

    gs.ecs.insert(map);

    gs.ecs
        .insert(Point::new(player_start_pos.x, player_start_pos.y));

    // for i in 0..10 {
    //     gs.ecs
    //         .create_entity()
    //         .with(Position { x: i * 7, y: 20 })
    //         .with(Renderable {
    //             glyph: to_cp437('Ä'),
    //             fg: RGB::named(vga::RED_BRIGHT),
    //             bg: RGB::named(vga::BLACK),
    //         })
    //         .with(LeftMover {})
    //         .build();
    // }

    // let descent_sounds: Vec<Sound> = (20..100)
    //     .rev()
    //     .flat_map(|x| {
    //         (1..10).rev().flat_map(move |y| {
    //             vec![Sound {
    //                 sound_type: SoundType::Tone(x * y * y),
    //                 duration: Duration::from_millis((y as f64 / 1.5) as u64),
    //             }]
    //         })
    //     })
    //     .collect();

    // let descent_effect = gs.sound_system.render_sound_effect(SoundEffect {
    //     sounds: descent_sounds,
    // });

    // let _ = gs.sound_system.play_sound(descent_effect);

    // let effect = gs.sound_system.render_sound_effect(SoundEffect {
    //     sounds: vec![Sound {
    //         sound_type: SoundType::Tone(3500),
    //         duration: Duration::from_millis(4000),
    //     }],
    // });

    // let _ = gs.sound_system.play_sound(effect);

    main_loop(context, gs)
}
